# -*- coding: utf-8 -*-
# Copyright (c) 2014-2022 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Collection of serial interfaces to OLED devices.
"""

# Example usage:
#
#   from luma.core.interface.serial import i2c, spi
#   from luma.core.render import canvas
#   from luma.oled.device import ssd1306, sh1106
#   from PIL import ImageDraw
#
#   serial = i2c(port=1, address=0x3C)
#   device = ssd1306(serial)
#
#   with canvas(device) as draw:
#      draw.rectangle(device.bounding_box, outline="white", fill="black")
#      draw.text(30, 40, "Hello World", fill="white")
#
# As soon as the with-block scope level is complete, the graphics primitives
# will be flushed to the device.
#
# Creating a new canvas is effectively 'carte blanche': If you want to retain
# an existing canvas, then make a reference like:
#
#    c = canvas(device)
#    for X in ...:
#        with c as draw:
#            draw.rectangle(...)
#
# As before, as soon as the with block completes, the canvas buffer is flushed
# to the device

from time import sleep
from luma.core.device import device, parallel_device
from luma.core.virtual import character
from luma.oled.device.color import color_device
from luma.oled.device.greyscale import greyscale_device
import luma.core.error
from luma.core.framebuffer import full_frame
from luma.core.bitmap_font import embedded_fonts
import luma.oled.const
from luma.oled.device.framebuffer_mixin import __framebuffer_mixin

__all__ = ["ssd1306", "ssd1309", "ssd1322", "ssd1362", "ssd1322_nhd", "ssd1325", "ssd1327", "ssd1331", "ssd1351", "sh1106", "ws0010", "winstar_weh"]


class sh1106(device):
    """
    Serial interface to a monochrome SH1106 OLED display.

    On creation, an initialization sequence is pumped to the display
    to properly configure it. Further control commands can then be called to
    affect the brightness and other settings.
    """

    def __init__(self, serial_interface=None, width=128, height=64, rotate=0, **kwargs):
        super(sh1106, self).__init__(luma.oled.const.sh1106, serial_interface)
        self.capabilities(width, height, rotate)
        self._pages = self._h // 8

        settings = {
            (128, 128): dict(multiplex=0xFF, displayoffset=0x02),
            (128, 64): dict(multiplex=0x3F, displayoffset=0x00),
            (128, 32): dict(multiplex=0x20, displayoffset=0x0F)
        }.get((width, height))

        if settings is None:
            raise luma.core.error.DeviceDisplayModeError(
                f"Unsupported display mode: {width} x {height}")

        self.command(
            self._const.DISPLAYOFF,
            self._const.MEMORYMODE,
            self._const.SETHIGHCOLUMN,      0xB0, 0xC8,
            self._const.SETLOWCOLUMN,       0x10, 0x40,
            self._const.SETSEGMENTREMAP,
            self._const.NORMALDISPLAY,
            self._const.SETMULTIPLEX,       settings['multiplex'],
            self._const.DISPLAYALLON_RESUME,
            self._const.SETDISPLAYOFFSET,   settings['displayoffset'],
            self._const.SETDISPLAYCLOCKDIV, 0xF0,
            self._const.SETPRECHARGE,       0x22,
            self._const.SETCOMPINS,         0x12,
            self._const.SETVCOMDETECT,      0x20,
            self._const.CHARGEPUMP,         0x14)

        self.contrast(0x7F)
        self.clear()
        self.show()

    def display(self, image):
        """
        Takes a 1-bit :py:mod:`PIL.Image` and dumps it to the SH1106
        OLED display.

        :param image: Image to display.
        :type image: :py:mod:`PIL.Image`
        """
        assert image.mode == self.mode
        assert image.size == self.size

        image = self.preprocess(image)

        set_page_address = 0xB0
        image_data = image.getdata()
        pixels_per_page = self.width * 8
        buf = bytearray(self.width)

        for y in range(0, int(self._pages * pixels_per_page), pixels_per_page):
            self.command(set_page_address, 0x02, 0x10)
            set_page_address += 1
            offsets = [y + self.width * i for i in range(8)]

            for x in range(self.width):
                buf[x] = \
                    (image_data[x + offsets[0]] and 0x01) | \
                    (image_data[x + offsets[1]] and 0x02) | \
                    (image_data[x + offsets[2]] and 0x04) | \
                    (image_data[x + offsets[3]] and 0x08) | \
                    (image_data[x + offsets[4]] and 0x10) | \
                    (image_data[x + offsets[5]] and 0x20) | \
                    (image_data[x + offsets[6]] and 0x40) | \
                    (image_data[x + offsets[7]] and 0x80)

            self.data(list(buf))


class ssd1306(device):
    """
    Serial interface to a monochrome SSD1306 OLED display.

    On creation, an initialization sequence is pumped to the display
    to properly configure it. Further control commands can then be called to
    affect the brightness and other settings.

    :param serial_interface: The serial interface (usually a
        :py:class:`luma.core.interface.serial.i2c` instance) to delegate sending
        data and commands through.
    :param width: The number of horizontal pixels (optional, defaults to 128).
    :type width: int
    :param height: The number of vertical pixels (optional, defaults to 64).
    :type height: int
    :param rotate: An integer value of 0 (default), 1, 2 or 3 only, where 0 is
        no rotation, 1 is rotate 90° clockwise, 2 is 180° rotation and 3
        represents 270° rotation.
    :type rotate: int
    """

    def __init__(self, serial_interface=None, width=128, height=64, rotate=0, **kwargs):
        super(ssd1306, self).__init__(luma.oled.const.ssd1306, serial_interface)
        self.capabilities(width, height, rotate)

        # Supported modes
        settings = {
            (128, 64): dict(multiplex=0x3F, displayclockdiv=0x80, compins=0x12),
            (128, 32): dict(multiplex=0x1F, displayclockdiv=0x80, compins=0x02),
            (96, 16): dict(multiplex=0x0F, displayclockdiv=0x60, compins=0x02),
            (64, 48): dict(multiplex=0x2F, displayclockdiv=0x80, compins=0x12),
            (64, 32): dict(multiplex=0x1F, displayclockdiv=0x80, compins=0x12)
        }.get((width, height))

        if settings is None:
            raise luma.core.error.DeviceDisplayModeError(
                f"Unsupported display mode: {width} x {height}")

        self._pages = height // 8
        self._mask = [1 << (i // width) % 8 for i in range(width * height)]
        self._offsets = [(width * (i // (width * 8))) + (i % width) for i in range(width * height)]
        self._colstart = (0x80 - self._w) // 2
        self._colend = self._colstart + self._w

        self.command(
            self._const.DISPLAYOFF,
            self._const.SETDISPLAYCLOCKDIV, settings['displayclockdiv'],
            self._const.SETMULTIPLEX,       settings['multiplex'],
            self._const.SETDISPLAYOFFSET,   0x00,
            self._const.SETSTARTLINE,
            self._const.CHARGEPUMP,         0x14,
            self._const.MEMORYMODE,         0x00,
            self._const.SETSEGMENTREMAP,
            self._const.COMSCANDEC,
            self._const.SETCOMPINS,         settings['compins'],
            self._const.SETPRECHARGE,       0xF1,
            self._const.SETVCOMDETECT,      0x40,
            self._const.DISPLAYALLON_RESUME,
            self._const.NORMALDISPLAY)

        self.contrast(0xCF)
        self.clear()
        self.show()

    def display(self, image):
        """
        Takes a 1-bit :py:mod:`PIL.Image` and dumps it to the OLED
        display.

        :param image: Image to display.
        :type image: :py:mod:`PIL.Image`
        """
        assert image.mode == self.mode
        assert image.size == self.size

        image = self.preprocess(image)

        self.command(
            # Column start/end address
            self._const.COLUMNADDR, self._colstart, self._colend - 1,
            # Page start/end address
            self._const.PAGEADDR, 0x00, self._pages - 1)

        buf = bytearray(self._w * self._pages)
        off = self._offsets
        mask = self._mask

        idx = 0
        for pix in image.getdata():
            if pix > 0:
                buf[off[idx]] |= mask[idx]
            idx += 1

        self.data(list(buf))


class ssd1309(ssd1306):
    """
    Serial interface to a monochrome SSD1309 OLED display.

    On creation, an initialization sequence is pumped to the display
    to properly configure it. Further control commands can then be called to
    affect the brightness and other settings.

    :param serial_interface: The serial interface (usually a
        :py:class:`luma.core.interface.serial.spi` instance) to delegate sending
        data and commands through.
    :param width: The number of horizontal pixels (optional, defaults to 128).
    :type width: int
    :param height: The number of vertical pixels (optional, defaults to 64).
    :type height: int
    :param rotate: An integer value of 0 (default), 1, 2 or 3 only, where 0 is
        no rotation, 1 is rotate 90° clockwise, 2 is 180° rotation and 3
        represents 270° rotation.
    :type rotate: int

    .. versionadded:: 3.1.0
    """


class ssd1331(color_device):
    """
    Serial interface to a 16-bit color (5-6-5 RGB) SSD1331 OLED display.

    On creation, an initialization sequence is pumped to the display to
    properly configure it. Further control commands can then be called to
    affect the brightness and other settings.

    :param serial_interface: The serial interface (usually a
        :py:class:`luma.core.interface.serial.spi` instance) to delegate
        sending data and commands through.
    :param width: The number of horizontal pixels (optional, defaults to 96).
    :type width: int
    :param height: The number of vertical pixels (optional, defaults to 64).
    :type height: int
    :param rotate: An integer value of 0 (default), 1, 2 or 3 only, where 0 is
        no rotation, 1 is rotate 90° clockwise, 2 is 180° rotation and 3
        represents 270° rotation.
    :type rotate: int
    :param framebuffer: Framebuffering strategy, currently instances of
        ``diff_to_previous`` or ``full_frame`` are only supported.
    :type framebuffer: str
    """

    def __init__(self, serial_interface=None, width=96, height=64, rotate=0,
                 framebuffer=None, **kwargs):
        super(ssd1331, self).__init__(serial_interface, width, height, rotate,
                                      framebuffer, **kwargs)

    def _supported_dimensions(self):
        return [(96, 64)]

    def _init_sequence(self):
        self.command(
            0xAE,        # Display off
            0xA0, 0x72,  # Seg remap
            0xA1, 0x00,  # Set Display start line
            0xA2, 0x00,  # Set display offset
            0xA4,        # Normal display
            0xA8, 0x3F,  # Set multiplex
            0xAD, 0x8E,  # Master configure
            0xB0, 0x0B,  # Power save mode
            0xB1, 0x74,  # Phase12 period
            0xB3, 0xD0,  # Clock divider
            0x8A, 0x80,  # Set precharge speed A
            0x8B, 0x80,  # Set precharge speed B
            0x8C, 0x80,  # Set precharge speed C
            0xBB, 0x3E,  # Set pre-charge voltage
            0xBE, 0x3E,  # Set voltage
            0x87, 0x0F)  # Master current control

    def _set_position(self, top, right, bottom, left):
        self.command(
            0x15, left, right - 1,    # Set column addr
            0x75, top, bottom - 1)    # Set row addr

    def contrast(self, level):
        """
        Switches the display contrast to the desired level, in the range
        0-255. Note that setting the level to a low (or zero) value will
        not necessarily dim the display to nearly off. In other words,
        this method is **NOT** suitable for fade-in/out animation.

        :param level: Desired contrast level in the range of 0-255.
        :type level: int
        """
        assert 0 <= level <= 255
        self.command(0x81, level,  # Set contrast A
                     0x82, level,  # Set contrast B
                     0x83, level)  # Set contrast C


class ssd1351(color_device):
    """
    Serial interface to the 16-bit color (5-6-5 RGB) SSD1351 OLED display.

    On creation, an initialization sequence is pumped to
    the display to properly configure it. Further control commands can then be
    called to affect the brightness and other settings.

    :param serial_interface: The serial interface (usually a
        :py:class:`luma.core.interface.serial.spi` instance) to delegate
        sending data and commands through.
    :param width: The number of horizontal pixels (optional, defaults to 128).
    :type width: int
    :param height: The number of vertical pixels (optional, defaults to 128).
    :type height: int
    :param rotate: An integer value of 0 (default), 1, 2 or 3 only, where 0 is
        no rotation, 1 is rotate 90° clockwise, 2 is 180° rotation and 3
        represents 270° rotation.
    :type rotate: int
    :param framebuffer: Framebuffering strategy, currently instances of
        ``diff_to_previous`` or ``full_frame`` are only supported.
    :type framebuffer: str
    :param bgr: Set to ``True`` if device pixels are BGR order (rather than RGB).
    :type bgr: bool
    :param h_offset: Horizontal offset (in pixels) of screen to device memory
        (default: 0)
    :type h_offset: int
    :param v_offset: Vertical offset (in pixels) of screen to device memory
        (default: 0)
    :type h_offset: int

    .. versionadded:: 2.3.0
    """

    def __init__(self, serial_interface=None, width=128, height=128, rotate=0,
                 framebuffer=None, h_offset=0, v_offset=0,
                 bgr=False, **kwargs):

        # RGB or BGR order
        self._color_order = 0x04 if bgr else 0x00

        if h_offset != 0 or v_offset != 0:
            def offset(bbox):
                left, top, right, bottom = bbox
                return (left + h_offset, top + v_offset, right + h_offset, bottom + v_offset)
            self._apply_offsets = offset

        super(ssd1351, self).__init__(serial_interface, width, height, rotate, framebuffer, **kwargs)

    def _supported_dimensions(self):
        return [(96, 96), (128, 128), (128, 96)]

    def _init_sequence(self):
        self.command(0xFD, 0x12)               # Unlock IC MCU interface
        self.command(0xFD, 0xB1)               # Command A2,B1,B3,BB,BE,C1 accessible if in unlock state
        self.command(0xAE)                     # Display off
        self.command(0xB3, 0xF1)               # Clock divider
        self.command(0xCA, 0x7F)               # Mux ratio
        self.command(0x15, 0x00, self.width - 1)    # Set column address
        self.command(0x75, 0x00, self.height - 1)   # Set row address
        self.command(0xA0, 0x70 | self._color_order)  # Segment remapping
        self.command(0xA1, 0x00)               # Set Display start line
        self.command(0xA2, 0x00)               # Set display offset
        self.command(0xB5, 0x00)               # Set GPIO
        self.command(0xAB, 0x01)               # Function select (internal - diode drop)
        self.command(0xB1, 0x32)               # Precharge
        self.command(0xB4, 0xA0, 0xB5, 0x55)   # Set segment low voltage
        self.command(0xBE, 0x05)               # Set VcomH voltage
        self.command(0xC7, 0x0F)               # Contrast master
        self.command(0xB6, 0x01)               # Precharge2
        self.command(0xA6)                     # Normal display

    def _set_position(self, top, right, bottom, left):
        self.command(0x15, left, right - 1)    # Set column addr
        self.command(0x75, top, bottom - 1)    # Set row addr
        self.command(0x5C)                     # Write RAM

    def contrast(self, level):
        """
        Switches the display contrast to the desired level, in the range
        0-255. Note that setting the level to a low (or zero) value will
        not necessarily dim the display to nearly off. In other words,
        this method is **NOT** suitable for fade-in/out animation.

        :param level: Desired contrast level in the range of 0-255.
        :type level: int
        """
        assert 0 <= level <= 255
        self.command(0xC1, level, level, level)

    def command(self, cmd, *args):
        """
        Sends a command and an (optional) sequence of arguments through to the
        delegated serial interface. Note that the arguments are passed through
        as data.
        """
        self._serial_interface.command(cmd)
        if len(args) > 0:
            self._serial_interface.data(list(args))


class ssd1322(greyscale_device):
    """
    Serial interface to a 4-bit greyscale SSD1322 OLED display.

    On creation, an initialization sequence is pumped to the
    display to properly configure it. Further control commands can then be
    called to affect the brightness and other settings.

    :param serial_interface: The serial interface (usually a
       :py:class:`luma.core.interface.serial.spi` instance) to delegate sending
       data and commands through.
    :param width: The number of horizontal pixels (optional, defaults to 96).
    :type width: int
    :param height: The number of vertical pixels (optional, defaults to 64).
    :type height: int
    :param rotate: An integer value of 0 (default), 1, 2 or 3 only, where 0 is
        no rotation, 1 is rotate 90° clockwise, 2 is 180° rotation and 3
        represents 270° rotation.
    :type rotate: int
    :param mode: Supplying "1" or "RGB" effects a different rendering
         mechanism, either to monochrome or 4-bit greyscale.
    :type mode: str
    :param framebuffer: Framebuffering strategy, currently instances of
        ``diff_to_previous`` or ``full_frame`` are only supported
    :type framebuffer: str

    """

    def __init__(self, serial_interface=None, width=256, height=64, rotate=0,
                 mode="RGB", framebuffer=None, **kwargs):
        self._column_offset = (480 - width) // 2
        super(ssd1322, self).__init__(luma.oled.const.ssd1322, serial_interface,
                                      width, height, rotate, mode, framebuffer,
                                      nibble_order=0, **kwargs)

    def _supported_dimensions(self):
        return [(256, 64), (256, 48), (256, 32),
                (128, 64), (128, 48), (128, 32),
                (64, 64),  (64, 48),  (64, 32)]

    def _init_sequence(self):
        self.command(0xFD, 0x12)        # Unlock IC
        self.command(0xA4)              # Display off (all pixels off)
        self.command(0xB3, 0xF2)        # Display divide clockratio/freq
        self.command(0xCA, 0x3F)        # Set MUX ratio
        self.command(0xA2, 0x00)        # Display offset
        self.command(0xA1, 0x00)        # Display start Line
        self.command(0xA0, 0x14, 0x11)  # Set remap & dual COM Line
        self.command(0xB5, 0x00)        # Set GPIO (disabled)
        self.command(0xAB, 0x01)        # Function select (internal Vdd)
        self.command(0xB4, 0xA0, 0xFD)  # Display enhancement A (External VSL)
        self.command(0xC7, 0x0F)        # Master contrast (reset)
        self.command(0xB9)              # Set default greyscale table
        self.command(0xB1, 0xF0)        # Phase length
        self.command(0xD1, 0x82, 0x20)  # Display enhancement B (reset)
        self.command(0xBB, 0x0D)        # Pre-charge voltage
        self.command(0xB6, 0x08)        # 2nd precharge period
        self.command(0xBE, 0x00)        # Set VcomH
        self.command(0xA6)              # Normal display (reset)
        self.command(0xA9)              # Exit partial display

    def _set_position(self, top, right, bottom, left):
        width = right - left
        pix_start = self._column_offset + left
        coladdr_start = pix_start >> 2
        coladdr_end = (pix_start + width >> 2) - 1

        self.command(0x15, coladdr_start, coladdr_end)  # set column addr
        self.command(0x75, top, bottom - 1)             # Reset row addr
        self.command(0x5C)                              # Enable MCU to write data into RAM

    def command(self, cmd, *args):
        """
        Sends a command and an (optional) sequence of arguments through to the
        delegated serial interface. Note that the arguments are passed through
        as data.
        """
        self._serial_interface.command(cmd)
        if len(args) > 0:
            self._serial_interface.data(list(args))


class ssd1362(greyscale_device):
    """
    Serial interface to a 4-bit greyscale SSD1362 OLED display.

    On creation, an initialization sequence is pumped to the
    display to properly configure it. Further control commands can then be
    called to affect the brightness and other settings.

    :param serial_interface: The serial interface (usually a
       :py:class:`luma.core.interface.serial.spi` instance) to delegate sending
       data and commands through.
    :param width: The number of horizontal pixels (optional, defaults to 96).
    :type width: int
    :param height: The number of vertical pixels (optional, defaults to 64).
    :type height: int
    :param rotate: An integer value of 0 (default), 1, 2 or 3 only, where 0 is
        no rotation, 1 is rotate 90° clockwise, 2 is 180° rotation and 3
        represents 270° rotation.
    :type rotate: int
    :param mode: Supplying "1" or "RGB" effects a different rendering
         mechanism, either to monochrome or 4-bit greyscale.
    :type mode: str
    :param framebuffer: Framebuffering strategy, currently instances of
        ``diff_to_previous`` or ``full_frame`` are only supported
    :type framebuffer: str

    .. versionadded:: 3.4.0
    """

    def __init__(self, serial_interface=None, width=256, height=64, rotate=0,
                 mode="RGB", framebuffer=None, **kwargs):
        super(ssd1362, self).__init__(luma.oled.const.ssd1362, serial_interface,
                                      width, height, rotate, mode, framebuffer,
                                      nibble_order=1, **kwargs)

    def _supported_dimensions(self):
        return [(256, 64)]

    def _init_sequence(self):
        self.command(
            0xAB,              # Set Vdd Mode
            0x01,              # ELW2106AA VCI = 3.0V
            0xAD, 0x9E,        # Set IREF selection
            0x15, 0x00, 0x7F,  # Set column address
            0x75, 0x00, 0x3F,  # Set row address
            0xA0, 0x43,        # Set Re-map
            0xA1, 0x00,        # Set display start line
            0xA2, 0x00,        # Set display offset
            0xA4,              # Set display mode
            0xA8, 0x3F,        # Set multiplex ratio
            0xB1, 0x11,        # Set Phase1,2 length
            0xB3, 0xF0,        # Set display clock divide ratio
            0xB9,              # Grey scale table
            0xBC, 0x04,        # Set pre-charge voltage
            0xBE, 0x05)        # Set VCOMH deselect level, 0.82 * Vcc

    def _set_position(self, top, right, bottom, left):
        self.command(
            0x15, left >> 1, (right - 1) >> 1,  # set column addr
            0x75, top, bottom - 1)              # set row addr


class ssd1322_nhd(greyscale_device):
    """Similar to ssd1322 but several options are hard coded: width, height and
    frame buffer"""

    def __init__(self, serial_interface=None, width=128, height=64, rotate=0,
                 mode="RGB", framebuffer=full_frame(), **kwargs):
        super(ssd1322_nhd, self).__init__(luma.oled.const.ssd1322, serial_interface,
                                      128, 64, rotate, mode, framebuffer,
                                      nibble_order=0, **kwargs)

    def _supported_dimensions(self):
        return [(128, 64)]

    def _init_sequence(self):
        self.command(0xFD, 0x12)  # Unlock IC
        self.command(0xAE)  # Display off
        self.command(0xB3, 0x91)  # Display divide clockratio/freq
        self.command(0xCA, 0x3F)  # Set MUX ratio
        self.command(0xA2, 0x00)  # Display offset
        self.command(0xAB, 0x01)  # Internal VDD
        self.command(0xA0, 0x16, 0x11)  # Set remap & dual COM Line
        self.command(0xC7, 0x0F)  # Master contrast (reset)
        self.command(0xC1, 0x9F)  # Set contrast current
        self.command(0xB1, 0xF2)  # Phase 1 and 2 lengths
        self.command(0xBB, 0x1F)  # Pre-charge voltage
        self.command(0xB4, 0xA0, 0xFD)  # Display enhancement A (External VSL)
        self.command(0xBE, 0x04)  # Set VcomH
        self.command(0xA6)  # Normal display (reset)
        self.command(0xAF)  # Exit partial display

    def _set_position(self, top, bottom):
        coladdr_start = 28
        coladdr_end = 91

        self.command(0x15, coladdr_start, coladdr_end)  # set column addr
        self.command(0x75, top, bottom - 1)             # Reset row addr
        self.command(0x5C)                              # Enable MCU to write data into RAM

    def command(self, cmd, *args):
        """
        Sends a command and an (optional) sequence of arguments through to the
        delegated serial interface. Note that the arguments are passed through
        as data.
        """
        self._serial_interface.command(cmd)
        if len(args) > 0:
            self._serial_interface.data(list(args))

    def _render_mono(self, buf, pixel_data):
        i = 0
        for pix in pixel_data:
            if pix > 0:
                buf[i] = 0xFF

            i += 1

    def _render_greyscale(self, buf, pixel_data):
        i = 0
        for r, g, b in pixel_data:
            # RGB->Greyscale luma calculation into 4-bits
            grey = (r * 306 + g * 601 + b * 117) >> 14

            # NHD uses 2 SEG lines and one COM line per pixel
            if grey > 0:
                buf[i // 2] |= (grey << 4) | grey
            i += 2

    def display(self, image):
        """
        Takes a 1-bit monochrome or 24-bit RGB image and renders it
        to the greyscale OLED display. RGB pixels are converted to 8-bit
        greyscale values using a simplified Luma calculation, based on
        *Y'=0.299R'+0.587G'+0.114B'*.

        :param image: the image to render
        :type image: PIL.Image.Image
        """
        assert image.mode == self.mode
        assert image.size == self.size

        image = self.preprocess(image)

        for image, bounding_box in self.framebuffer.redraw(image):
            left, top, right, bottom = bounding_box
            width = right - left
            height = bottom - top

            buf = bytearray(width * height)
            self._set_position(top, bottom)
            self._populate(buf, image.getdata())
            self.data(list(buf))


class ssd1325(greyscale_device):
    """
    Serial interface to a 4-bit greyscale SSD1325 OLED display.

    On creation, an initialization sequence is pumped to the
    display to properly configure it. Further control commands can then be
    called to affect the brightness and other settings.
    """

    def __init__(self, serial_interface=None, width=128, height=64, rotate=0,
                 mode="RGB", framebuffer=None, **kwargs):
        super(ssd1325, self).__init__(luma.core.const.common, serial_interface,
                                      width, height, rotate, mode, framebuffer,
                                      nibble_order=1, **kwargs)

    def _supported_dimensions(self):
        return [(128, 64)]

    def _init_sequence(self):
        self.command(
            0xAE,               # Diplay off (all pixels off)
            0xB3, 0xF2,         # Display divide clockratio/freq
            0xA8, 0x3F,         # Set MUX ratio
            0xA2, 0x4C,         # Display offset
            0xA1, 0x00,         # Display start line
            0xAD, 0x02,         # Master configuration (external Vcc)
            0xA0, 0x50,         # Set remap (enable COM remap & split odd/even)
            0x86,               # Set current range (full)
            0xB8, 0x01, 0x11,   # Set greyscale table
            0x22, 0x32, 0x43,   # .. cont
            0x54, 0x65, 0x76,   # .. cont
            0xB2, 0x51,         # Set row period
            0xB1, 0x55,         # Set phase length
            0xB4, 0x03,         # Set pre-charge compensation level
            0xB0, 0x28,         # Set pre-charge compensation enable
            0xBC, 0x01,         # Pre-charge voltage
            0xBE, 0x00,         # Set VcomH
            0xBF, 0x02,         # Set VSL (not connected)
            0xA4)               # Normal dislay

    def _set_position(self, top, right, bottom, left):
        self.command(
            0x15, left >> 1, (right - 1) >> 1,  # set column addr
            0x75, top, bottom - 1)  # set row addr


class ssd1327(greyscale_device):
    """
    Serial interface to a 4-bit greyscale SSD1327 OLED display.

    On creation, an initialization sequence is pumped to the display to
    properly configure it. Further control commands can then be called to
    affect the brightness and other settings.

    .. versionadded:: 2.4.0
    """

    def __init__(self, serial_interface=None, width=128, height=128, rotate=0,
                 mode="RGB", framebuffer=None, **kwargs):
        super(ssd1327, self).__init__(luma.core.const.common, serial_interface,
                                      width, height, rotate, mode, framebuffer,
                                      nibble_order=1, **kwargs)

    def _supported_dimensions(self):
        return [(128, 128)]

    def _init_sequence(self):
        self.command(
            0xAE,               # Display off (all pixels off)
            0xA0, 0x53,         # Segment remap (com split, com remap, nibble remap, column remap)
            0xA1, 0x00,         # Display start line
            0xA2, 0x00,         # Display offset
            0xA4,               # regular display
            0xA8, 0x7F)         # set multiplex ratio: 127

        self.command(
            0xB8, 0x01, 0x11,   # Set greyscale table
            0x22, 0x32, 0x43,   # .. cont
            0x54, 0x65, 0x76)   # .. cont

        self.command(
            0xB3, 0x00,         # Front clock divider: 0, Fosc: 0
            0xAB, 0x01,         # Enable Internal Vdd

            0xB1, 0xF1,         # Set phase periods - 1: 1 clk, 2: 15 clks
            0xBC, 0x08,         # Pre-charge voltage: Vcomh
            0xBE, 0x07,         # COM deselect voltage level: 0.86 x Vcc

            0xD5, 0x62,         # Enable 2nd pre-charge
            0xB6, 0x0F)         # 2nd Pre-charge period: 15 clks

    def _set_position(self, top, right, bottom, left):
        self.command(
            0x15, left >> 1, (right - 1) >> 1,  # set column addr
            0x75, top, bottom - 1)  # set row addr


class ws0010(parallel_device, character, __framebuffer_mixin):
    """
    Serial interface to a monochrome Winstar WS0010 OLED display.  This
    interface will work with most ws0010 powered devices including the weg010016.

    :param serial_interface: The serial interface (usually a
        :py:class:`luma.core.interface.serial.parallel` instance) to delegate sending
        data and commands through.
    :param width: The number of pixels laid out horizontally.
    :type width: int
    :param height: The number of pixels laid out vertically.
    :type height: int
    :param undefined: The character to display if the font doesn't contain
        the requested character
    :type undefined: str
    :param font: Allows you to override the internal font by passing in an alternate
    :type font: :py:mod:`PIL.ImageFont`
    :param selected_font: Select one of the ws0010's embedded font tables (see
        note).  Can be selected by number or name.  Default is 'FT00'
        (English Japanese).
    :type selected_font: int or str
    :param exec_time: Time in seconds to wait for a command to complete.
        Default is 50 μs (1e-6 * 50) which is enough for all ws0010 commands.
    :type exec_time: float
    :param rotate: An integer value of 0 (default), 1, 2 or 3 only, where 0 is
        no rotation, 1 is rotate 90° clockwise, 2 is 180° rotation and 3
        represents 270° rotation.
    :type rotate: int
    :param framebuffer: Framebuffering strategy, currently values of
        ``diff_to_previous`` or ``full_frame`` are only supported.
    :type framebuffer: str

    To place text on the display, simply assign the text to the 'text'
    instance variable::

        p = parallel(RS=7, E=8, PINS=[25,24,23,18])
        my_display = ws0010(p, selected_font='FT01')
        my_display.text = 'WS0010 Display\\nFont FT01 5x8'

    For more details on how to use the 'text' interface see
    :class:`luma.core.virtual.character`

    .. note:
        The ws0010 is a fully graphical device that also supports character-based
        operations similar to LCD displays such as the hd44780.  This driver
        uses the graphics mode but includes functionality to allow the device
        to act like a character-based one.  To do this it includes four
        embedded fonts which are available in two sizes (5x8 pixel and 5x10
        pixel).  You can select which font to use by providing the appropriate
        name or number to selected_font during initialization.  You can also use
        any PIL.ImageFont object instead by providing it to the font parameter.
        If you do, the internal fonts will be bypassed.

        Available Internal Fonts
        +--------+---------+----------------------+------+
        | Number | Name    | Font                 | Size |
        +--------+---------+----------------------+------+
        |    0   | FT00    | English Japanese     | 5x8  |
        +--------+---------+----------------------+------+
        |    1   | FT01    | Western European I   | 5x8  |
        +--------+---------+----------------------+------+
        |    2   | FT10    | English Russian      | 5x8  |
        +--------+---------+----------------------+------+
        |    3   | FT11    | Western European II  | 5x8  |
        +--------+---------+----------------------+------+
        |    4   | FT00_10 | English Japanese     | 5x10 |
        +--------+---------+----------------------+------+
        |    5   | FT01_10 | Western European I   | 5x10 |
        +--------+---------+----------------------+------+
        |    6   | FT10_10 | English Russian      | 5x10 |
        +--------+---------+----------------------+------+
        |    7   | FT11_10 | Western European II  | 5x10 |
        +--------+---------+----------------------+------+

    .. versionadded:: 3.6.0
    """

    def __init__(self, serial_interface=None, width=100, height=16, undefined='_', font=None,
                 selected_font=0, exec_time=1e-6 * 50, rotate=0, framebuffer=None,
                 const=luma.oled.const.ws0010, **kwargs):
        super(ws0010, self).__init__(const, serial_interface, exec_time=exec_time, **kwargs)
        self.capabilities(width, height, rotate)
        self.init_framebuffer(framebuffer)
        self.font = font if font is not None else embedded_fonts(self._const.FONTDATA, selected_font=selected_font)
        self._undefined = undefined
        self.device = self

        # Supported modes
        supported = (width, height) in [(40, 8), (40, 16), (60, 8), (60, 16), (80, 8), (80, 16), (100, 8), (100, 16)]
        if not supported:
            raise luma.core.error.DeviceDisplayModeError(
                f"Unsupported display mode: {width} x {height}")

        # In case display just powered up, sleep to be sure it has finished
        # its internal initialization
        sleep(0.5)
        self._reset()
        self.text = ""

    def _reset(self):
        """
        WS0010 Initialization Routine

        Reset (not really needed after power-on but does no harm and is useful
        in case the display is already initialized)

        Send five 0s in four bit mode (e.g. d4-d7 are the only pins that matter)
        Set interface data length FUNCTIONSET|DL8 or FUNCTIONSET|DL4 (again in 4 bit mode)
        Turn Display Off
        Turn internal power off
        Configure Entry Mode (Set Cursor to Right and Shift Off)
        Set Graphics Mode and Internal Power On
        Turn Display On (with Cursor Off and Blink Off)

        Device is now ready to receive data
        """

        dl = 0x03 if self._bitmode == 8 else 0x02
        self.command(0x00, 0x00, dl, self._const.FUNCTIONSET | (dl << 4))
        self.command(self._const.DISPLAYOFF)  # Set Display Off
        self.command(self._const.POWEROFF)
        self.command(self._const.ENTRY)  # Set entry mode to direction right, no shift
        self.command(self._const.POWERON | self._const.GRAPHIC)  # Turn internal power on and set into graphics mode
        self.command(self._const.DISPLAYON)  # Turn Display back on

    def display(self, image):
        """
        Takes a 1-bit :py:mod:`PIL.Image` and dumps it to the ws0010
        OLED display.
        """
        assert image.mode == self.mode
        assert image.size == self.size

        image = self.preprocess(image)

        for _, bounding_box in self.framebuffer.redraw(image):
            # Expand bounding box to align to cell height boundary (8)
            # TODO: Should consider whether this should be moved into framebuffer class
            left, top, right, bottom = bounding_box
            top = top // 8 * 8
            bottom = bottom // 8 * 8 if not bottom % 8 else (bottom // 8 + 1) * 8

            if self._bitmode == 4:
                # If in 4 bit mode, issue reset to make sure we are in sync with the display
                self._reset()

            w = right - left
            h = bottom - top
            mask = [1 << (i // w) % 8 for i in range(w * h)]
            off = [(w * (i // (w * 8))) + (i % w) for i in range(w * h)]
            buf = bytearray(w * h // 8)

            image_segment = image.crop((left, top, right, bottom))
            for idx, pix in enumerate(image_segment.getdata()):
                if pix > 0:
                    buf[off[idx]] |= mask[idx]

            lines = int((bottom - top) / 8)
            lineSize = int(len(buf) / lines)
            for i in range(lines):
                self.command(self._const.DDRAMADDR + left, self._const.CGRAMADDR + i + (top // 8))  # Set display to current line at the starting column to update
                self.data(buf[lineSize * i:lineSize * (i + 1)])   # Send section of current line that needs to be changed

    def get_font(self, ft):
        """
        Load one of the devices embedded fonts by its index value or name and
        return it

        :param val: The index or the name of the font to return
        :type val: int or str
        """

        return self.font.load(ft)


class winstar_weh(ws0010):
    """
    Serial interface to a monochrome Winstar WEH OLED display.  This is the
    character version of the display using the ws0010 controller.  This class
    provides the same ``text`` property as the ws0010 interface so you can set
    the text value which will be rendered to the display's screen.  This
    interface uses a variant of the ws0010 controller's built-in font that is
    designed to match the grid structure of the weh displays (see note below).

    :param serial_interface: The serial interface (usually a
        :py:class:`luma.core.interface.serial.parallel` instance) to delegate sending
        data and commands through.
    :param width: The number of characters that can be displayed on a single line.
        Example: the weh001602a has a width of 16 characters.
    :type width: int
    :param height: The number of lines the display has.  Example: the weh001602a
        has a height of 2 lines.
    :type height: int
    :param undefined: The character to display if the font doesn't contain
        the requested character
    :type undefined: str
    :param font: Allows you to override the internal font by passing in an alternate
    :type font: :py:mod:`PIL.ImageFont`
    :param default_table: Select one of the ws0010's four embedded font tables
        (see :py:class:`ws0010` documentation)
    :type default_table: int
    :param embedded_font: Select the size of the embedded font to use.  Allowed
        sizes are 5x8 (default) and 5x10
    :type embedded_font: str '5x8' or '5x10'
    :param exec_time: Time in seconds to wait for a command to complete.
        Default is 50 μs (1e-6 * 50)
    :type exec_time: float
    :param rotate: An integer value of 0 (default), 1, 2 or 3 only, where 0 is
        no rotation, 1 is rotate 90° clockwise, 2 is 180° rotation and 3
        represents 270° rotation.
    :type rotate: int
    :param framebuffer: Framebuffering strategy, currently values of
        ``diff_to_previous`` or ``full_frame`` are only supported.
    :type framebuffer: str

    .. note:
      The WEH devices mimic character displays by having a small gap every fifth
      horizontal pixel. So, while the device can be addressed as a graphical
      display, the images that are shown will have a one pixel gap every 5th pixel.
      For this reason, you should be careful when designing you screens to
      account for this effect.

      The included text functionality automatically takes care of this for you
      if you are displaying character-based content.

    .. versionadded:: 3.6.0
    """

    def __init__(self, serial_interface=None, width=16, height=2, **kwargs):
        super(winstar_weh, self).__init__(const=luma.oled.const.winstar_weh, serial_interface=serial_interface, width=width * 5, height=height * 8, xwidth=5, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014-18 Richard Hull and contributors
# See LICENSE.rst for details.

from luma.oled.device import sh1106
from luma.core.render import canvas

from baseline_data import get_reference_data, primitives
from helpers import serial, assert_invalid_dimensions, setup_function  # noqa: F401
from unittest.mock import Mock, call


def test_init_128x64():
    """
    SH1106 OLED with a 128 x 64 resolution works correctly.
    """
    sh1106(serial)
    serial.command.assert_has_calls([
        # Initial burst are initialization commands
        call(174, 32, 16, 176, 200, 0, 16, 64, 161, 166, 168, 63, 164,
             211, 0, 213, 240, 217, 34, 218, 18, 219, 32, 141, 20),
        # set contrast
        call(129, 127),
        # reset the display
        call(176, 2, 16),
        call(177, 2, 16),
        call(178, 2, 16),
        call(179, 2, 16),
        call(180, 2, 16),
        call(181, 2, 16),
        call(182, 2, 16),
        call(183, 2, 16)
    ])

    # Next 1024 are all data: zero's to clear the RAM
    # (1024 = 128 * 64 / 8)
    serial.data.assert_has_calls([call([0] * 128)] * 8)


def test_init_invalid_dimensions():
    """
    SH1106 OLED with an invalid resolution raises a
    :py:class:`luma.core.error.DeviceDisplayModeError`.
    """
    assert_invalid_dimensions(sh1106, serial, 77, 105)


def test_display():
    """
    SH1106 OLED screen can draw and display an image.
    """
    device = sh1106(serial)
    serial.reset_mock()

    recordings = []

    def data(data):
        recordings.append({'data': data})

    def command(*cmd):
        recordings.append({'command': list(cmd)})

    serial.command = Mock(side_effect=command, unsafe=True)
    serial.data = Mock(side_effect=data, unsafe=True)

    # Use the same drawing primitives as the demo
    with canvas(device) as draw:
        primitives(device, draw)

    serial.data.assert_called()
    serial.command.assert_called()

    # To regenerate test data, uncomment the following (remember not to commit though)
    # ================================================================================
    # from baseline_data import save_reference_data
    # save_reference_data("demo_sh1106", recordings)

    assert recordings == get_reference_data('demo_sh1106')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014-18 Richard Hull and contributors
# See LICENSE.rst for details.

import pytest

from luma.oled.device import ssd1331
from luma.core.render import canvas
from luma.core.framebuffer import full_frame

from baseline_data import get_reference_data, primitives
from helpers import serial, assert_invalid_dimensions, setup_function  # noqa: F401
from unittest.mock import call


def test_init_96x64():
    """
    SSD1331 OLED with a 96 x 64 resolution works correctly.
    """
    ssd1331(serial, framebuffer=full_frame())
    serial.command.assert_has_calls([
        # Initial burst are initialization commands
        call(174, 160, 114, 161, 0, 162, 0, 164, 168, 63, 173,
             142, 176, 11, 177, 116, 179, 208, 138, 128, 139,
             128, 140, 128, 187, 62, 190, 62, 135, 15),
        # set contrast
        call(129, 255, 130, 255, 131, 255),
        # reset the display
        call(21, 0, 95, 117, 0, 63),
        # called last, is a command to show the screen
        call(175)
    ])

    # Next 12288 are all data: zero's to clear the RAM
    # (12288 = 96 * 64 * 2)
    serial.data.assert_called_once_with([0] * 96 * 64 * 2)


def test_init_invalid_dimensions():
    """
    SSD1331 OLED with an invalid resolution raises a
    :py:class:`luma.core.error.DeviceDisplayModeError`.
    """
    assert_invalid_dimensions(ssd1331, serial, 23, 57)


def test_hide():
    """
    SSD1331 OLED screen content can be hidden.
    """
    device = ssd1331(serial, framebuffer=full_frame())
    serial.reset_mock()
    device.hide()
    serial.command.assert_called_once_with(174)


def test_show():
    """
    SSD1331 OLED screen content can be displayed.
    """
    device = ssd1331(serial, framebuffer=full_frame())
    serial.reset_mock()
    device.show()
    serial.command.assert_called_once_with(175)


def test_display():
    """
    SSD1331 OLED screen can draw and display an image.
    """
    device = ssd1331(serial, framebuffer=full_frame())
    serial.reset_mock()

    # Use the same drawing primitives as the demo
    with canvas(device) as draw:
        primitives(device, draw)

    # Initial command to reset the display
    serial.command.assert_called_once_with(21, 0, 95, 117, 0, 63)

    # To regenerate test data, uncomment the following (remember not to commit though)
    # ================================================================================
    # from baseline_data import save_reference_data
    # save_reference_data("demo_ssd1331", serial.data.call_args.args[0])

    # Next 12288 bytes are data representing the drawn image
    serial.data.assert_called_once_with(get_reference_data('demo_ssd1331'))


@pytest.mark.parametrize("bit,expected_16_bit_color", [
    (7, [0b10000100, 0b00010000]),
    (6, [0b01000010, 0b00001000]),
    (5, [0b00100001, 0b00000100]),
    (4, [0b00010000, 0b10000010]),
    (3, [0b00001000, 0b01000001]),
    (2, [0b00000000, 0b00100000]),
    (1, [0b00000000, 0b00000000]),
    (0, [0b00000000, 0b00000000]),
])
def test_16bit_rgb_packing(bit, expected_16_bit_color):
    """
    Checks that 8 bit red component is packed into first 5 bits
    Checks that 8 bit green component is packed into next 6 bits
    Checks that 8 bit blue component is packed into remaining 5 bits
    """
    device = ssd1331(serial, framebuffer=full_frame())
    serial.reset_mock()

    rgb_color = (2 ** bit,) * 3
    expected = expected_16_bit_color * device.width * device.height

    with canvas(device) as draw:
        draw.rectangle(device.bounding_box, outline=rgb_color, fill=rgb_color)

    serial.data.assert_called_once_with(expected)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Richard Hull and contributors
# See LICENSE.rst for details.

from luma.oled.device import ssd1322_nhd
from luma.core.render import canvas
from luma.core.framebuffer import full_frame

from baseline_data import get_reference_data, primitives
from helpers import serial


def test_init_128x64():
    """
    SSD1322_NHD OLED with a 128 x 64 resolution works correctly.
    """
    recordings = []

    def data(data):
        recordings.append({'data': data})

    def command(*cmd):
        recordings.append({'command': list(cmd)})

    serial.command.side_effect = command
    serial.data.side_effect = data

    ssd1322_nhd(serial)

    assert serial.data.called
    assert serial.command.called

    assert recordings == [
        {'command': [253]}, {'data': [18]},
        {'command': [174]},
        {'command': [179]}, {'data': [145]},
        {'command': [202]}, {'data': [63]},
        {'command': [162]}, {'data': [0]},
        {'command': [171]}, {'data': [1]},
        {'command': [160]}, {'data': [22, 17]},
        {'command': [199]}, {'data': [15]},
        {'command': [193]}, {'data': [159]},
        {'command': [177]}, {'data': [242]},
        {'command': [187]}, {'data': [31]},
        {'command': [180]}, {'data': [160, 253]},
        {'command': [190]}, {'data': [4]},
        {'command': [166]},
        {'command': [175]},
        {'command': [193]}, {'data': [127]},
        {'command': [21]}, {'data': [28, 91]},
        {'command': [117]}, {'data': [0, 63]},
        {'command': [92]}, {'data': [0] * (128 * 64)},
        {'command': [175]}
    ]


def test_hide():
    """
    SSD1322_NHD OLED screen content can be hidden.
    """
    device = ssd1322_nhd(serial)
    serial.reset_mock()
    device.hide()
    serial.command.assert_called_once_with(174)


def test_show():
    """
    SSD1322_NHD OLED screen content can be displayed.
    """
    device = ssd1322_nhd(serial)
    serial.reset_mock()
    device.show()
    serial.command.assert_called_once_with(175)


def test_greyscale_display():
    """
    SSD1322_NHD OLED screen can draw and display a greyscale image.
    """
    device = ssd1322_nhd(serial, mode="RGB", framebuffer=full_frame())
    serial.reset_mock()

    recordings = []

    def data(data):
        recordings.append({'data': data})

    def command(*cmd):
        recordings.append({'command': list(cmd)})

    serial.command.side_effect = command
    serial.data.side_effect = data

    # Use the same drawing primitives as the demo
    with canvas(device) as draw:
        primitives(device, draw)

    assert serial.data.called
    assert serial.command.called

    # To regenerate test data, uncomment the following (remember not to commit though)
    # ================================================================================
    # from baseline_data import save_reference_data
    # save_reference_data("demo_ssd1322_nhd_greyscale", recordings)

    assert recordings == get_reference_data('demo_ssd1322_nhd_greyscale')


def test_monochrome_display():
    """
    SSD1322_NHD OLED screen can draw and display a monochrome image.
    """
    device = ssd1322_nhd(serial, mode="1", framebuffer=full_frame())
    serial.reset_mock()

    recordings = []

    def data(data):
        recordings.append({'data': data})

    def command(*cmd):
        recordings.append({'command': list(cmd)})

    serial.command.side_effect = command
    serial.data.side_effect = data

    # Use the same drawing primitives as the demo
    with canvas(device) as draw:
        primitives(device, draw)

    assert serial.data.called
    assert serial.command.called

    # To regenerate test data, uncomment the following (remember not to commit though)
    # ================================================================================
    # from baseline_data import save_reference_data
    # save_reference_data("demo_ssd1322_nhd_monochrome", recordings)

    assert recordings == get_reference_data('demo_ssd1322_nhd_monochrome')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014-18 Richard Hull and contributors
# See LICENSE.rst for details.

from luma.oled.device import ssd1309
from luma.core.render import canvas

from baseline_data import primitives, get_reference_data
from helpers import serial, assert_invalid_dimensions, setup_function  # noqa: F401
from unittest.mock import call


def test_init_128x64():
    """
    SSD1309 OLED with a 128 x 64 resolution works correctly.
    """
    ssd1309(serial)
    serial.command.assert_has_calls([
        # Initial burst are initialization commands
        call(174, 213, 128, 168, 63, 211, 0, 64, 141, 20, 32, 0,
             161, 200, 218, 18, 217, 241, 219, 64, 164, 166),
        # set contrast
        call(129, 207),
        # reset the display
        call(33, 0, 127, 34, 0, 7),
        # called last, is a command to show the screen
        call(175)
    ])

    # Next are all data: zero's to clear the RAM
    serial.data.assert_called_once_with([0] * (128 * 64 // 8))


def test_init_invalid_dimensions():
    """
    SSD1309 OLED with an invalid resolution raises a
    :py:class:`luma.core.error.DeviceDisplayModeError`.
    """
    assert_invalid_dimensions(ssd1309, serial, 59, 22)


def test_hide():
    """
    SSD1309 OLED screen content can be hidden.
    """
    device = ssd1309(serial)
    serial.reset_mock()
    device.hide()
    serial.command.assert_called_once_with(174)


def test_show():
    """
    SSD1309 OLED screen content can be displayed.
    """
    device = ssd1309(serial)
    serial.reset_mock()
    device.show()
    serial.command.assert_called_once_with(175)


def test_display():
    """
    SSD1309 OLED screen can draw and display an image.
    """
    device = ssd1309(serial)
    serial.reset_mock()

    # Use the same drawing primitives as the demo
    with canvas(device) as draw:
        primitives(device, draw)

    # Initial command to reset the display
    serial.command.assert_called_once_with(33, 0, 127, 34, 0, 7)

    # To regenerate test data, uncomment the following (remember not to commit though)
    # ================================================================================
    # from baseline_data import save_reference_data
    # save_reference_data("demo_ssd1309", serial.data.call_args.args[0])

    # Next 1024 bytes are data representing the drawn image
    serial.data.assert_called_once_with(get_reference_data('demo_ssd1309'))



# -*- coding: utf-8 -*-
# Copyright (c) 2014-2022 Richard Hull and contributors
# See LICENSE.rst for details.

"""
OLED display driver for SSD1306, SSD1309, SSD1322, SSD1325, SSD1327, SSD1331,
SSD1351, SSD1362 and SH1106 devices.
"""

__version__ = '3.10.0'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020-2022 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Tests for the :py:class:`luma.oled.device.ws0010` device.
"""

from luma.core.framebuffer import full_frame
from luma.oled.device import ws0010, winstar_weh
from luma.core.render import canvas
from luma.core.util import bytes_to_nibbles

from PIL import Image, ImageDraw
from unittest.mock import Mock, call

CLEAR = 0x01
HOME = 0x02
ENTRY = 0x06
DISPLAYOFF = 0x08
DISPLAYON = 0x0C
POWEROFF = 0x13
POWERON = 0x17
GRAPHIC = 0x08
CHAR = 0x00
FUNCTIONSET = 0x29
DL8 = 0x10
DL4 = 0x00
DDRAMADDR = 0x80
CGRAMADDR = 0x40

interface = Mock(unsafe=True, _bitmode=4, _pulse_time=1e-6)


def test_init_4bitmode():
    ws0010(interface, framebuffer=full_frame())

    to_4 = \
        [call(0, 0, 0, 0, 0, 2, 2, 9)]

    calls = \
        to_4 + \
        [call(*bytes_to_nibbles([DISPLAYOFF]))] + \
        [call(*bytes_to_nibbles([POWEROFF]))] + \
        [call(*bytes_to_nibbles([ENTRY]))] + \
        [call(*bytes_to_nibbles([POWERON | GRAPHIC]))] + \
        [call(*bytes_to_nibbles([DISPLAYON]))] + \
        [call(*bytes_to_nibbles([DDRAMADDR, CGRAMADDR]))] + \
        [call(*bytes_to_nibbles([DDRAMADDR, CGRAMADDR + 1]))]

    interface.command.assert_has_calls(calls)

    # Data to clear the screen
    calls = \
        [call(bytes_to_nibbles([0] * 100))] + \
        [call(bytes_to_nibbles([0] * 100))]

    interface.data.assert_has_calls(calls)


def test_init_8bitmode():
    interface._bitmode = 8
    ws0010(interface, framebuffer=full_frame())

    to_4 = \
        [call(0, 0, 3, 57)]

    calls = \
        to_4 + \
        [call(DISPLAYOFF)] + \
        [call(POWEROFF)] + \
        [call(ENTRY)] + \
        [call(POWERON | GRAPHIC)] + \
        [call(DISPLAYON)] + \
        [call(*[DDRAMADDR, CGRAMADDR])] + \
        [call(*[DDRAMADDR, CGRAMADDR + 1])]

    interface.command.assert_has_calls(calls)

    # Data to clear the screen
    calls = \
        [call(bytearray([0] * 100))] + \
        [call(bytearray([0] * 100))]

    interface.data.assert_has_calls(calls)


def test_display():
    interface._bitmode = 4
    d = ws0010(interface, framebuffer=full_frame())
    interface.reset_mock()

    # Use canvas to create an all white screen
    with canvas(d) as drw:
        drw.rectangle((0, 0, 99, 15), fill='white', outline='white')

    line1 = [call.command(*bytes_to_nibbles([DDRAMADDR, CGRAMADDR])),
        call.data([0x0F] * 200)]
    line2 = [call.command(*bytes_to_nibbles([DDRAMADDR, CGRAMADDR + 1])),
        call.data([0x0F] * 200)]

    interface.assert_has_calls(line1 + line2)


def test_get_font():
    interface._bitmode = 8
    device = ws0010(interface, framebuffer=full_frame())

    img = Image.new('1', (11, 8), 0)
    device.font.current = 0
    f0 = device.font.current
    f3 = device.get_font('FT11')
    drw = ImageDraw.Draw(img)

    assert f0.getsize('\u00E0') == (5, 8)
    assert f0.getsize('A\u00E0') == (11, 8)

    drw.text((0, 0), '\u00E0', font=f0, fill='white')  # Font 0: α
    drw.text((6, 0), '\u00E0', font=f3, fill='white')  # Font 3: à

    # Compare to image containing αà
    assert img.tobytes() == \
        b'\x01\x00\x00\x80I\xc0\xa8 \x91\xe0\x92 i\xe0\x00\x00'


def test_unsupported_display_mode():
    import luma.core
    try:
        ws0010(interface, width=99, height=15)
    except luma.core.error.DeviceDisplayModeError as ex:
        assert str(ex) == "Unsupported display mode: 99 x 15"


def test_winstar_weh():

    class image_retaining_framebuffer(full_frame):

        def redraw(self, image):
            self.image = image
            yield image, (0, 0) + image.size

    interface._bitmode = 4
    device = winstar_weh(interface, framebuffer=image_retaining_framebuffer())

    device.text = 'This is a test'

    img = Image.new('1', (80, 16), 0)
    drw = ImageDraw.Draw(img)
    drw.text((0, 0), 'This is a test', font=device.font, fill='white')

    assert 'This is a test' == str(device.text)
    assert device.font.getbbox('This is a test') == (0, 0, 70, 8)
    assert device.framebuffer.image == img



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014-18 Richard Hull and contributors
# See LICENSE.rst for details.

from luma.oled.device import ssd1322
from luma.core.render import canvas
from luma.core.framebuffer import full_frame

from baseline_data import get_reference_data, primitives
from helpers import serial, assert_invalid_dimensions, setup_function  # noqa: F401


def test_init_256x64():
    """
    SSD1322 OLED with a 256 x 64 resolution works correctly.
    """
    recordings = []

    def data(data):
        recordings.append({'data': data})

    def command(*cmd):
        recordings.append({'command': list(cmd)})

    serial.command.side_effect = command
    serial.data.side_effect = data

    ssd1322(serial)

    assert serial.data.called
    assert serial.command.called

    assert recordings == [
        {'command': [253]}, {'data': [18]},
        {'command': [164]},
        {'command': [179]}, {'data': [242]},
        {'command': [202]}, {'data': [63]},
        {'command': [162]}, {'data': [0]},
        {'command': [161]}, {'data': [0]},
        {'command': [160]}, {'data': [20, 17]},
        {'command': [181]}, {'data': [0]},
        {'command': [171]}, {'data': [1]},
        {'command': [180]}, {'data': [160, 253]},
        {'command': [199]}, {'data': [15]},
        {'command': [185]},
        {'command': [177]}, {'data': [240]},
        {'command': [209]}, {'data': [130, 32]},
        {'command': [187]}, {'data': [13]},
        {'command': [182]}, {'data': [8]},
        {'command': [190]}, {'data': [0]},
        {'command': [166]},
        {'command': [169]},
        {'command': [193]}, {'data': [127]},
        {'command': [21]}, {'data': [28, 91]},
        {'command': [117]}, {'data': [0, 63]},
        {'command': [92]}, {'data': [0] * (256 * 64 // 2)},
        {'command': [175]}
    ]


def test_init_invalid_dimensions():
    """
    SSD1322 OLED with an invalid resolution raises a
    :py:class:`luma.core.error.DeviceDisplayModeError`.
    """
    assert_invalid_dimensions(ssd1322, serial, 128, 77)


def test_hide():
    """
    SSD1322 OLED screen content can be hidden.
    """
    device = ssd1322(serial)
    serial.reset_mock()
    device.hide()
    serial.command.assert_called_once_with(174)


def test_show():
    """
    SSD1322 OLED screen content can be displayed.
    """
    device = ssd1322(serial)
    serial.reset_mock()
    device.show()
    serial.command.assert_called_once_with(175)


def test_greyscale_display():
    """
    SSD1322 OLED screen can draw and display a greyscale image.
    """
    device = ssd1322(serial, mode="RGB", framebuffer=full_frame())
    serial.reset_mock()

    recordings = []

    def data(data):
        recordings.append({'data': data})

    def command(*cmd):
        recordings.append({'command': list(cmd)})

    serial.command.side_effect = command
    serial.data.side_effect = data

    # Use the same drawing primitives as the demo
    with canvas(device) as draw:
        primitives(device, draw)

    assert serial.data.called
    assert serial.command.called

    # To regenerate test data, uncomment the following (remember not to commit though)
    # ================================================================================
    # from baseline_data import save_reference_data
    # save_reference_data("demo_ssd1322_greyscale", recordings)

    assert recordings == get_reference_data('demo_ssd1322_greyscale')


def test_monochrome_display():
    """
    SSD1322 OLED screen can draw and display a monochrome image.
    """
    device = ssd1322(serial, mode="1", framebuffer=full_frame())
    serial.reset_mock()

    recordings = []

    def data(data):
        recordings.append({'data': data})

    def command(*cmd):
        recordings.append({'command': list(cmd)})

    serial.command.side_effect = command
    serial.data.side_effect = data

    # Use the same drawing primitives as the demo
    with canvas(device) as draw:
        primitives(device, draw)

    assert serial.data.called
    assert serial.command.called

    # To regenerate test data, uncomment the following (remember not to commit though)
    # ================================================================================
    # from baseline_data import save_reference_data
    # save_reference_data("demo_ssd1322_monochrome", recordings)

    assert recordings == get_reference_data('demo_ssd1322_monochrome')



# -*- coding: utf-8 -*-
# Copyright (c) 2018-2022 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Base class for SSD13xx greyscale devices

.. versionadded:: 3.0.0
"""

from abc import abstractmethod, ABCMeta

from luma.core.device import device
import luma.core.error
import luma.oled.const
from luma.oled.device.framebuffer_mixin import __framebuffer_mixin


class greyscale_device(device, __framebuffer_mixin):
    __metaclass__ = ABCMeta

    def __init__(self, const, serial_interface, width, height, rotate, mode,
                 framebuffer, nibble_order, **kwargs):
        super(greyscale_device, self).__init__(const, serial_interface)
        self.capabilities(width, height, rotate, mode)
        self.init_framebuffer(framebuffer)

        self._populate = self._render_mono if mode == "1" else self._render_greyscale
        self._nibble_order = nibble_order

        if (width, height) not in self._supported_dimensions():
            raise luma.core.error.DeviceDisplayModeError(
                f"Unsupported display mode: {width} x {height}")

        self._init_sequence()
        self.contrast(0x7F)
        self.clear()
        self.show()

    @abstractmethod
    def _supported_dimensions(self):
        """
        Enumerates the screen resolutions that the specific device supports, as
        a list of tuples; e.g.: ``[(96, 64), (96, 32), (96, 16)]``
        """
        pass  # pragma: no cover

    @abstractmethod
    def _init_sequence(self):
        """
        Concrete implementations should call the initiation sequence for the
        specific device. Invoked from the constructor, but no assumptions should
        be made about what has been initialized so far. No return value is
        expected.
        """
        pass  # pragma: no cover

    @abstractmethod
    def _set_position(self, top, right, bottom, left):
        """
        Invoked once as part of the devices display refresh. The four coordinates
        form a bounding box that determines the area of the screen that will get
        get redrawn; thus the concrete implementations should send the correct
        command sequence to the device to set that bounding box. No return value
        is expected.
        """
        pass  # pragma: no cover

    def _render_mono(self, buf, pixel_data):
        i = 0
        nibble_order = self._nibble_order
        for pix in pixel_data:
            if pix > 0:
                if i % 2 == nibble_order:
                    buf[i // 2] |= 0xF0
                else:
                    buf[i // 2] |= 0x0F

            i += 1

    def _render_greyscale(self, buf, pixel_data):
        i = 0
        nibble_order = self._nibble_order
        for r, g, b in pixel_data:
            # RGB->Greyscale luma calculation into 4-bits
            grey = (r * 306 + g * 601 + b * 117) >> 14

            if grey > 0:
                if i % 2 == nibble_order:
                    buf[i // 2] |= (grey << 4)
                else:
                    buf[i // 2] |= grey

            i += 1

    def _inflate_bbox(self, bounding_box):
        """
        Realign the left and right edges of the bounding box such that they are
        inflated to align modulo 4.

        this method is optional, and used mainly to accommodate devices with
        COM/SEG GDDRAM structures that store pixels in 4-bit nibbles.
        """
        left, top, right, bottom = bounding_box
        return (
            left & 0xFFFC,
            top,
            right if right % 4 == 0 else (right & 0xFFFC) + 0x04,
            bottom)

    def display(self, image):
        """
        Takes a 1-bit monochrome or 24-bit RGB image and renders it
        to the greyscale OLED display. RGB pixels are converted to 4-bit
        greyscale values using a simplified Luma calculation, based on
        *Y'=0.299R'+0.587G'+0.114B'*.

        :param image: The image to render.
        :type image: PIL.Image.Image
        """
        assert image.mode == self.mode
        assert image.size == self.size

        image = self.preprocess(image)

        for _, bounding_box in self.framebuffer.redraw(image):
            left, top, right, bottom = self._inflate_bbox(bounding_box)
            cropped_image_segment = image.crop((left, top, right, bottom))
            width = right - left
            height = bottom - top

            buf = bytearray(width * height >> 1)
            self._set_position(top, right, bottom, left)
            self._populate(buf, cropped_image_segment.getdata())
            self.data(list(buf))



# -*- coding: utf-8 -*-
# Copyright (c) 2018-2022 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Base class for SSD13xx color devices

.. versionadded:: 3.0.0
"""

from abc import abstractmethod, ABCMeta

from luma.core.device import device
import luma.core.error
import luma.core.framebuffer
import luma.oled.const
from luma.oled.device.framebuffer_mixin import __framebuffer_mixin


class color_device(device, __framebuffer_mixin):
    __metaclass__ = ABCMeta

    def __init__(self, serial_interface, width, height, rotate, framebuffer, **kwargs):
        super(color_device, self).__init__(luma.oled.const.common, serial_interface)
        self.capabilities(width, height, rotate, mode="RGB")
        self.init_framebuffer(framebuffer)

        if (width, height) not in self._supported_dimensions():
            raise luma.core.error.DeviceDisplayModeError(
                f"Unsupported display mode: {width} x {height}")

        self._init_sequence()
        self.contrast(0xFF)
        self.clear()
        self.show()

    @abstractmethod
    def _supported_dimensions(self):
        """
        Enumerates the screen resolutions that the specific device supports, as
        a list of tuples; e.g.: ``[(96, 64), (96, 32), (96, 16)]``
        """
        pass  # pragma: no cover

    @abstractmethod
    def _init_sequence(self):
        """
        Concrete implementations should call the initiation sequence for the
        specific device. Invoked from the constructor, but no assumptions should
        be made about what has been initialized so far. No return value is
        expected.
        """
        pass  # pragma: no cover

    @abstractmethod
    def _set_position(self, top, right, bottom, left):
        """
        Invoked once as part of the devices display refresh. The four coordinates
        form a bounding box that determines the area of the screen that will get
        get redrawn; thus the concrete implementations should send the correct
        command sequence to the device to set that bounding box. No return value
        is expected.
        """
        pass  # pragma: no cover

    def _apply_offsets(self, bbox):
        return bbox

    def display(self, image):
        """
        Renders a 24-bit RGB image to the Color OLED display.

        :param image: The image to render.
        :type image: PIL.Image.Image
        """
        assert image.mode == self.mode
        assert image.size == self.size

        image = self.preprocess(image)

        for image, bounding_box in self.framebuffer.redraw(image):
            left, top, right, bottom = self._apply_offsets(bounding_box)
            width = right - left
            height = bottom - top

            self._set_position(top, right, bottom, left)

            i = 0
            buf = bytearray(width * height * 2)
            for r, g, b in image.getdata():
                if not r == g == b == 0:
                    # 65K format 1
                    buf[i] = r & 0xF8 | g >> 5
                    buf[i + 1] = g << 3 & 0xE0 | b >> 3
                i += 2

            self.data(list(buf))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014-18 Richard Hull and contributors
# See LICENSE.rst for details.

from luma.oled.device import ssd1325
from luma.core.render import canvas
from luma.core.framebuffer import full_frame, diff_to_previous

from baseline_data import get_reference_data, primitives
from helpers import serial, assert_invalid_dimensions, setup_function  # noqa: F401
from unittest.mock import call


def test_init_128x64():
    """
    SSD1325 OLED with a 128 x 64 resolution works correctly.
    """
    ssd1325(serial, framebuffer=full_frame())
    serial.command.assert_has_calls([
        call(174, 179, 242, 168, 63, 162, 76, 161, 0, 173, 2, 160, 80, 134, 184,
             1, 17, 34, 50, 67, 84, 101, 118, 178, 81, 177, 85, 180, 3, 176, 40,
             188, 1, 190, 0, 191, 2, 164),
        call(129, 127),
        call(21, 0, 63, 117, 0, 63),
        call(175)
    ])

    # Next 4096 are all data: zero's to clear the RAM
    # (4096 = 128 * 64 / 2)
    serial.data.assert_called_once_with([0] * (128 * 64 // 2))


def test_init_invalid_dimensions():
    """
    SSD1325 OLED with an invalid resolution raises a
    :py:class:`luma.core.error.DeviceDisplayModeError`.
    """
    assert_invalid_dimensions(ssd1325, serial, 128, 77)


def test_hide():
    """
    SSD1325 OLED screen content can be hidden.
    """
    device = ssd1325(serial, framebuffer=full_frame())
    serial.reset_mock()
    device.hide()
    serial.command.assert_called_once_with(174)


def test_show():
    """
    SSD1325 OLED screen content can be displayed.
    """
    device = ssd1325(serial, framebuffer=full_frame())
    serial.reset_mock()
    device.show()
    serial.command.assert_called_once_with(175)


def test_greyscale_display():
    """
    SSD1325 OLED screen can draw and display a greyscale image.
    """
    device = ssd1325(serial, mode="RGB", framebuffer=full_frame())
    serial.reset_mock()

    # Use the same drawing primitives as the demo
    with canvas(device) as draw:
        primitives(device, draw)

    # Initial command to reset the display
    serial.command.assert_called_once_with(21, 0, 63, 117, 0, 63)

    # To regenerate test data, uncomment the following (remember not to commit though)
    # ================================================================================
    # from baseline_data import save_reference_data
    # save_reference_data("demo_ssd1325_greyscale", serial.data.call_args.args[0])

    # Next 4096 bytes are data representing the drawn image
    serial.data.assert_called_once_with(get_reference_data('demo_ssd1325_greyscale'))


def test_monochrome_display():
    """
    SSD1325 OLED screen can draw and display a monochrome image.
    """
    device = ssd1325(serial, mode="1", framebuffer=full_frame())
    serial.reset_mock()

    # Use the same drawing primitives as the demo
    with canvas(device) as draw:
        primitives(device, draw)

    # Initial command to reset the display
    serial.command.assert_called_once_with(21, 0, 63, 117, 0, 63)

    # To regenerate test data, uncomment the following (remember not to commit though)
    # ================================================================================
    # from baseline_data import save_reference_data
    # save_reference_data("demo_ssd1325_monochrome", serial.data.call_args.args[0])

    # Next 4096 bytes are data representing the drawn image
    serial.data.assert_called_once_with(get_reference_data('demo_ssd1325_monochrome'))


def test_framebuffer_override():
    """
    Reproduce https://github.com/rm-hull/luma.examples/issues/95
    """
    ssd1325(serial, mode="1", framebuffer=diff_to_previous())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-18 Richard Hull and contributors
# See LICENSE.rst for details.

import pytest

from luma.oled.device import ssd1351
from luma.core.render import canvas
from luma.core.framebuffer import full_frame

from baseline_data import get_reference_data, primitives
from helpers import serial, assert_invalid_dimensions, setup_function  # noqa: F401


def test_init_128x128():
    """
    SSD1351 OLED with a 128 x 128 resolution works correctly.
    """
    recordings = []

    def data(data):
        recordings.append({'data': data})

    def command(*cmd):
        recordings.append({'command': list(cmd)})

    serial.command.side_effect = command
    serial.data.side_effect = data

    ssd1351(serial, framebuffer=full_frame())

    assert serial.data.called
    assert serial.command.called

    assert recordings == [
        {'command': [253]}, {'data': [18]},
        {'command': [253]}, {'data': [177]},
        {'command': [174]},
        {'command': [179]}, {'data': [241]},
        {'command': [202]}, {'data': [127]},
        {'command': [21]}, {'data': [0, 127]},
        {'command': [117]}, {'data': [0, 127]},
        {'command': [160]}, {'data': [112]},
        {'command': [161]}, {'data': [0]},
        {'command': [162]}, {'data': [0]},
        {'command': [181]}, {'data': [0]},
        {'command': [171]}, {'data': [1]},
        {'command': [177]}, {'data': [50]},
        {'command': [180]}, {'data': [160, 181, 85]},
        {'command': [190]}, {'data': [5]},
        {'command': [199]}, {'data': [15]},
        {'command': [182]}, {'data': [1]},
        {'command': [166]},
        {'command': [193]}, {'data': [255, 255, 255]},
        {'command': [21]}, {'data': [0, 127]},
        {'command': [117]}, {'data': [0, 127]},
        {'command': [92]}, {'data': [0] * (128 * 128 * 2)},
        {'command': [175]}
    ]


def test_init_96x96_with_BGR():
    """
    SSD1351 OLED with a 96 x 96 resolution and BGR pixels works correctly.
    """
    recordings = []

    def data(data):
        recordings.append({'data': data})

    def command(*cmd):
        recordings.append({'command': list(cmd)})

    serial.command.side_effect = command
    serial.data.side_effect = data

    ssd1351(serial, width=96, height=96, bgr=True, framebuffer=full_frame())

    assert serial.data.called
    assert serial.command.called

    assert recordings == [
        {'command': [253]}, {'data': [18]},
        {'command': [253]}, {'data': [177]},
        {'command': [174]},
        {'command': [179]}, {'data': [241]},
        {'command': [202]}, {'data': [127]},
        {'command': [21]}, {'data': [0, 95]},
        {'command': [117]}, {'data': [0, 95]},
        {'command': [160]}, {'data': [116]},
        {'command': [161]}, {'data': [0]},
        {'command': [162]}, {'data': [0]},
        {'command': [181]}, {'data': [0]},
        {'command': [171]}, {'data': [1]},
        {'command': [177]}, {'data': [50]},
        {'command': [180]}, {'data': [160, 181, 85]},
        {'command': [190]}, {'data': [5]},
        {'command': [199]}, {'data': [15]},
        {'command': [182]}, {'data': [1]},
        {'command': [166]},
        {'command': [193]}, {'data': [255, 255, 255]},
        {'command': [21]}, {'data': [0, 95]},
        {'command': [117]}, {'data': [0, 95]},
        {'command': [92]}, {'data': [0] * (96 * 96 * 2)},
        {'command': [175]}
    ]


def test_offsets():
    """
    SSD1351 OLED with offsets works correctly.
    """
    recordings = []

    def data(data):
        recordings.append({'data': data})

    def command(*cmd):
        recordings.append({'command': list(cmd)})

    serial.command.side_effect = command
    serial.data.side_effect = data

    ssd1351(serial, width=96, height=96, h_offset=2, v_offset=1, framebuffer=full_frame())

    assert serial.data.called
    assert serial.command.called

    assert recordings == [
        {'command': [253]}, {'data': [18]},
        {'command': [253]}, {'data': [177]},
        {'command': [174]},
        {'command': [179]}, {'data': [241]},
        {'command': [202]}, {'data': [127]},
        {'command': [21]}, {'data': [0, 95]},
        {'command': [117]}, {'data': [0, 95]},
        {'command': [160]}, {'data': [112]},
        {'command': [161]}, {'data': [0]},
        {'command': [162]}, {'data': [0]},
        {'command': [181]}, {'data': [0]},
        {'command': [171]}, {'data': [1]},
        {'command': [177]}, {'data': [50]},
        {'command': [180]}, {'data': [160, 181, 85]},
        {'command': [190]}, {'data': [5]},
        {'command': [199]}, {'data': [15]},
        {'command': [182]}, {'data': [1]},
        {'command': [166]},
        {'command': [193]}, {'data': [255, 255, 255]},
        {'command': [21]}, {'data': [2, 97]},
        {'command': [117]}, {'data': [1, 96]},
        {'command': [92]}, {'data': [0] * (96 * 96 * 2)},
        {'command': [175]}
    ]


def test_init_invalid_dimensions():
    """
    SSD1351 OLED with an invalid resolution raises a
    :py:class:`luma.core.error.DeviceDisplayModeError`.
    """
    assert_invalid_dimensions(ssd1351, serial, 128, 77)


def test_hide():
    """
    SSD1351 OLED screen content can be hidden.
    """
    device = ssd1351(serial, framebuffer=full_frame())
    serial.reset_mock()
    device.hide()
    serial.command.assert_called_once_with(174)


def test_show():
    """
    SSD1351 OLED screen content can be displayed.
    """
    device = ssd1351(serial, framebuffer=full_frame())
    serial.reset_mock()
    device.show()
    serial.command.assert_called_once_with(175)


def test_display():
    """
    SSD1351 OLED screen can draw and display a color image.
    """
    device = ssd1351(serial, framebuffer=full_frame())
    serial.reset_mock()

    recordings = []

    def data(data):
        recordings.append({'data': data})

    def command(*cmd):
        recordings.append({'command': list(cmd)})

    serial.command.side_effect = command
    serial.data.side_effect = data

    # Use the same drawing primitives as the demo
    with canvas(device) as draw:
        primitives(device, draw)

    assert serial.data.called
    assert serial.command.called

    # To regenerate test data, uncomment the following (remember not to commit though)
    # ================================================================================
    # from baseline_data import save_reference_data
    # save_reference_data("demo_ssd1351", recordings)

    assert recordings == get_reference_data('demo_ssd1351')


@ pytest.mark.parametrize("bit,expected_16_bit_color", [
    (7, [0b10000100, 0b00010000]),
    (6, [0b01000010, 0b00001000]),
    (5, [0b00100001, 0b00000100]),
    (4, [0b00010000, 0b10000010]),
    (3, [0b00001000, 0b01000001]),
    (2, [0b00000000, 0b00100000]),
    (1, [0b00000000, 0b00000000]),
    (0, [0b00000000, 0b00000000]),
])
def test_16bit_rgb_packing(bit, expected_16_bit_color):
    """
    Checks that 8 bit red component is packed into first 5 bits
    Checks that 8 bit green component is packed into next 6 bits
    Checks that 8 bit blue component is packed into remaining 5 bits
    """
    device = ssd1351(serial, framebuffer=full_frame())
    serial.reset_mock()

    rgb_color = (2 ** bit,) * 3
    expected = expected_16_bit_color * device.width * device.height
    recordings = []

    def data(data):
        recordings.append({'data': data})

    def command(*cmd):
        recordings.append({'command': list(cmd)})

    serial.command.side_effect = command
    serial.data.side_effect = data

    with canvas(device) as draw:
        draw.rectangle(device.bounding_box, outline=rgb_color, fill=rgb_color)

    assert serial.data.called
    assert serial.command.called

    assert recordings == [
        {'command': [21]}, {'data': [0, 127]},
        {'command': [117]}, {'data': [0, 127]},
        {'command': [92]}, {'data': expected}
    ]



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Richard Hull and contributors
# See LICENSE.rst for details.

import luma.core.framebuffer


class __framebuffer_mixin(object):
    """
    Helper class for initializing the framebuffer. Its only purpose is to
    log a deprecation warning if a string framebuffer is specified.

    .. note::
        Specifying the framebuffer as a string will be removed at the next
        major release, and hence this mixin will become redundant and will
        also be removed at that point.

    .. versionadded:: 3.8.0
    """

    def init_framebuffer(self, framebuffer):
        if framebuffer is None:
            self.framebuffer = luma.core.framebuffer.diff_to_previous()
        elif isinstance(framebuffer, str):
            import warnings
            warnings.warn(
                "Specifying framebuffer as a string is now deprecated; Supply an instance of class full_frame() or diff_to_previous() instead",
                DeprecationWarning
            )
            self.framebuffer = getattr(luma.core.framebuffer, framebuffer)()
        else:
            self.framebuffer = framebuffer
